/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2002-2011 Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!

!------------------------------------------------------------------------------!
  MODULE small_box
!------------------------------------------------------------------------------!

      !  This module contains the basis vector of the small sub-cell 
      !  (small box) used for charge augmentation process

      USE kinds, ONLY : DP
!
      IMPLICIT NONE
      SAVE
      !  atb: lattice crystal vectors in "alatb" units (equivalent to "at")
      REAL(DP) :: atb(3,3) = RESHAPE( (/ 0.0_DP /), (/ 3, 3 /), (/ 0.0_DP /) )
      !  bgb: reciprocal lattice vectors, in 2pi/alatb units (equiv to "bg")
      REAL(DP) :: bgb(3,3) = RESHAPE( (/ 0.0_DP /), (/ 3, 3 /), (/ 0.0_DP /) )
      !  omegab: volume of the small boxes 
      REAl(DP) :: omegab = 0.0_DP
      !  alatb: lattice parameter of the boxes (the equivalent of "alat")
      REAL(DP) :: alatb  = 0.0_DP
      !  tpibab: 2pi/alatb
      REAL(DP) :: tpibab = 0.0_DP

!------------------------------------------------------------------------------!
   CONTAINS
!------------------------------------------------------------------------------!
!

     SUBROUTINE small_box_set( alat, omega, at, rat1, rat2, rat3, tprint )
       USE constants, ONLY: pi
       USE io_global, ONLY: stdout, ionode
       IMPLICIT NONE
       REAL(DP), INTENT(IN) :: alat, omega, at(3,3), rat1, rat2, rat3
       LOGICAL,  INTENT(IN) :: tprint

       alatb  = alat * rat1
       IF( alatb <= 0.0_DP ) CALL errore(' small_box_set ', ' alatb <= 0 ', 1 )
       tpibab = 2.0_DP * pi / alatb
       atb(:,1) = at(:,1)*alat * rat1 / alatb
       atb(:,2) = at(:,2)*alat * rat2 / alatb
       atb(:,3) = at(:,3)*alat * rat3 / alatb 
       omegab = omega * rat1 * rat2 * rat3
!
       CALL recips( atb(1,1), atb(1,2), atb(1,3), bgb(1,1), bgb(1,2), bgb(1,3) )

       IF( tprint .AND. ionode ) THEN
          WRITE( stdout,*)
          WRITE( stdout,220)
220       format( 3X, 'unit vectors of box grid cell',/,                        &
     &            3X, 'in real space:',25x,'in reciprocal space:')
          WRITE( stdout,'(3X,3f10.4,10x,3f10.4)') atb(:,1)*alatb, bgb(:,1)
          WRITE( stdout,'(3X,3f10.4,10x,3f10.4)') atb(:,2)*alatb, bgb(:,2)
          WRITE( stdout,'(3X,3f10.4,10x,3f10.4)') atb(:,3)*alatb, bgb(:,3)
       END IF

       RETURN
     END SUBROUTINE small_box_set
!
!------------------------------------------------------------------------------!
   END MODULE small_box
!------------------------------------------------------------------------------!
