/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2001 PWSCF group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!
!----------------------------------------------------------------------

subroutine dpsi_corr (evcq, psidvpsi_x, ik, ikq, nu)
  !-----------------------------------------------------------------------
  ! Used in the metallic case.
  ! If dpsi common variable contains the projection on the conduction
  ! states of the first variation of a wavefunction at a given k-point,
  ! this routine corrects dpsi in such a way that the density matrix
  ! is given by:   Sum_{k,nu} 2 * | dpsi > < psi |
  !
  USE kinds, only : DP
  use pwcom
  use phcom
  use d3com

  implicit none
  integer :: ik, ikq, nu, ibnd, jbnd
  ! index of the k-point under consideration
  ! index of the corresponding k+q point
  ! mode under consideration
  ! counter on bands
  ! counter on bands

  real (DP) :: wfshift, wgauss, w0gauss, deltae, wg1, wg2, wwg
  ! the shift coefficent for the wave function
  ! function computing the theta function
  ! function computing the derivative of theta
  ! difference of energy
  ! weight for metals
  ! weight for metals
  ! weight for metals

  complex (DP) :: evcq (npwx, nbnd), psidvpsi_x (nbnd, nbnd), &
       psidvpsi
  ! k+q point wavefunction
  ! < psi_{k+q} | V(q) | psi_k >
  !
  ! Multiplies dpsi by the theta function
  !
  do ibnd = 1, nbnd
     wg1 = wgauss ( (ef - et (ibnd, ik) ) / degauss, ngauss)
     call dscal (2 * npwq, wg1, dpsi (1, ibnd), 1)
  enddo
  !
  ! Adds to dpsi the term containing the valence wavefunctions
  !
  do ibnd = 1, nbnd
     do jbnd = 1, nbnd
        deltae = et (ibnd, ik) - et (jbnd, ikq)
        if (abs (deltae) .gt.1.0d-5) then
           wg1 = wgauss ( (ef - et (ibnd, ik) ) / degauss, ngauss)
           wg2 = wgauss ( (ef - et (jbnd, ikq) ) / degauss, ngauss)
           wwg = (wg1 - wg2) / deltae
        else
           wwg = - w0gauss ( (ef - et (ibnd, ik) ) / degauss, ngauss) &
                / degauss
        endif
        psidvpsi = 0.5d0 * wwg * psidvpsi_x (jbnd, ibnd)
        call zaxpy (npwq, psidvpsi, evcq (1, jbnd), 1, dpsi (1, ibnd), &
             1)
     enddo
  enddo
  !
  ! If necessary corrects dpsi with a term depending on FermiEnergy shift
  !
  if (ik.eq.ikq) then
     do ibnd = 1, nbnd_occ (ik)
        wfshift = 0.5d0 * ef_sh (nu) * w0gauss ( (ef - et (ibnd, ik) ) &
             / degauss, ngauss) / degauss
        call daxpy (2 * npw, wfshift, evcq (1, ibnd), 1, dpsi (1, ibnd) &
             , 1)
     enddo

  endif
  return

end subroutine dpsi_corr
