/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2007 Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!
!-----------------------------------------------------------------------
subroutine h_psi_meta (ldap, np, mp, psip, hpsi)
  !-----------------------------------------------------------------------
  !
  ! This routine computes the specific contribution from the meta-GGA
  ! potential to H*psi; the result is added to hpsi
  !
  USE kinds,     ONLY : DP
  USE cell_base, ONLY : tpiba
  USE lsda_mod,  ONLY : nspin, current_spin
  USE wvfct,     ONLY : igk, current_k
  USE gvecs,   ONLY : nls, nlsm
  USE gvect,     ONLY : g
  USE scf,       ONLY : kedtau
  USE klist,     ONLY : xk
  USE control_flags,        ONLY : gamma_only
  USE wavefunctions_module, ONLY : psic
  USE fft_base,             ONLY : dffts
  USE fft_interfaces,       ONLY : fwfft, invfft
  !
  implicit none
  !
  COMPLEX(DP), PARAMETER :: ci=(0.d0,1.d0)
  integer :: ldap, np, mp
  complex(DP) :: psip (ldap, mp), hpsi (ldap, mp)
  real (DP), allocatable :: kplusg (:)
!  complex (DP), allocatable :: psi(:)
  !
  integer :: im, j, nrxxs
  !
  CALL start_clock( 'h_psi_meta' )
  nrxxs = dffts%nnr
  allocate (kplusg(np))
  if (gamma_only) then
     !
     ! gamma algorithm
     !
     do im = 1, mp, 2
        do j =1,3
           psic(1:nrxxs) = ( 0.D0, 0.D0 )
           !
           kplusg (1:np) = (xk(j,current_k)+g(j,igk(1:np))) * tpiba
           if (im < mp ) then
              psic(nls(igk(1:np))) =  ci * kplusg(1:np) * &
                              ( psip (1:np,im) + ci * psip(1:np,im+1) )
              psic(nlsm(igk(1:np)))= -ci * kplusg(1:np) * &
                        CONJG ( psip (1:np,im) - ci * psip(1:np,im+1) )
           else
              psic(nls(igk(1:np))) =  ci * kplusg(1:np) *       psip(1:np,im) 
              psic(nlsm(igk(1:np)))= -ci * kplusg(1:np) * CONJG(psip(1:np,im))
           end if
           !
           CALL invfft ('Wave', psic, dffts)
           !
           psic(1:nrxxs) = kedtau(1:nrxxs,current_spin) * psic(1:nrxxs) 
           !
           CALL fwfft ('Wave', psic, dffts)
           !
           if ( im < mp ) then
              hpsi(1:np,im)  = hpsi(1:np,im)   - ci * kplusg(1:np) * 0.5d0 * &
                       ( psic(nls(igk(1:np))) + CONJG(psic(nlsm(igk(1:np)))) )
              hpsi(1:np,im+1)= hpsi(1:np,im+1) - kplusg(1:np) * 0.5d0 * &
                       ( psic(nls(igk(1:np))) - CONJG(psic(nlsm(igk(1:np)))) )
           else
              hpsi(1:np,im) = hpsi(1:np,im) - ci * kplusg(1:np) * &
                                                      psic(nls(igk(1:np)))
           end if
        end do
     end do
  else
     !
     ! generic k algorithm
     !
     do im = 1, mp
        do j =1,3
           psic(1:nrxxs) = ( 0.D0, 0.D0 )
           !
           kplusg (1:np) = (xk(j,current_k)+g(j,igk(1:np))) * tpiba
           psic(nls(igk(1:np))) = CMPLX(0d0, kplusg(1:np),kind=DP) * psip (1:np,im)
           !
           CALL invfft ('Wave', psic, dffts)
           !
           psic(1:nrxxs) = kedtau(1:nrxxs,current_spin) * psic(1:nrxxs) 
           !
           CALL fwfft ('Wave', psic, dffts)
           !
           hpsi(1:np,im) = hpsi(1:np,im) - &
                           CMPLX(0d0, kplusg(1:np),kind=DP) * psic(nls(igk(1:np)))
        end do
     end do
  end if
  deallocate (kplusg)
  CALL stop_clock( 'h_psi_meta' )


  return

end subroutine h_psi_meta
