/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2001-2008 Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!
!-----------------------------------------------------------------------
subroutine check_atoms (nvec, vec, trmat)
  !-----------------------------------------------------------------------
  !
  !     This routine tests that the atomic coordinates (or k-points)
  !     are different and not related by a lattice translation
  !
  !
  USE kinds
  implicit none
  !
  real(DP), parameter :: accep=1.d-5
  !
  integer, intent(in) :: nvec
  ! nvec : number of atomic positions (or k-points)
  real(DP), intent(in) :: vec (3, nvec), trmat (3, 3)
  ! vec  : cartesian coordinates of atomic positions (or k-points)
  ! trmat: transformation matrix to crystal axis
  !        ( = bg , basis of the real-space lattice, for atoms
  !          = at , basis of the rec.-space lattice, for k-points )
  !
  integer :: nv1, nv2
  real(DP), allocatable :: vaux(:,:)
  real(DP) :: zero (3) = 0.0_dp
  character(len=80) :: message
  logical, external :: eqvect
  !
  !   Copy input positions and transform them to crystal units
  !
  allocate ( vaux(3,nvec) )
  vaux = vec
  call cryst_to_cart ( nvec, vaux, trmat, -1)
  !
  !   Test that all the atomic positions (or k-points) are different
  !
  do nv1 = 1, nvec-1
     do nv2 = nv1+1, nvec
        if ( eqvect ( vaux (1,nv1), vaux (1,nv2), zero, accep ) ) then
           zero(:) = vaux (:,nv1)-vaux(:,nv2)
           IF ( ALL ( abs(zero) < accep ) ) THEN
              WRITE (message,'("atoms #",i4," and #",i4," overlap!")') nv1, nv2
           ELSE
              WRITE (message,'("atoms #",i4," and #",i4," differ by lattice &
                    &vector (",i2,",",i2,",",i2,") in crystal axis")') &
                    nv1, nv2, NINT(zero)
           ENDIF
           CALL errore ( 'check_atoms', TRIM(message), 1)
        end if
     enddo
  enddo
  !
  deallocate(vaux)
  return
end subroutine check_atoms

