/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

! Copyright (C) 2008 Dmitry Korotin dmitry@korotin.name, Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!


SUBROUTINE ortho_wfc(lda,ldb,wfc,ierr)
!This subroutine orthogonalizes wfcs.

  USE kinds,      ONLY : DP
  USE io_global,  ONLY : stdout
  USE mp_bands,   ONLY : intra_bgrp_comm
  USE mp,         ONLY : mp_sum
  USE noncollin_module, ONLY : noncolin, npol
  
  implicit none
  INTEGER, intent(in) :: lda,ldb
  INTEGER, intent(out) :: ierr
  COMPLEX(DP), intent(inout) :: wfc(lda,ldb)
  
  INTEGER :: i,j,k
  COMPLEX(DP), allocatable :: overlap(:,:),work(:,:), wfc_ortho(:,:)
  REAL(DP) , ALLOCATABLE :: e (:)

  ierr = 0

  ALLOCATE (overlap( lda , lda))    
  ALLOCATE (work   ( lda , lda))    
  ALLOCATE (e      ( lda))    
  ALLOCATE (wfc_ortho( lda , ldb))    
  !
  ! calculate overlap matrix
  !
  overlap = (0.d0,0.d0)
  work = (0.d0,0.d0)
  e = 0.d0

  CALL ZGEMM ('n', 'c', lda, lda, ldb, (1.d0, 0.d0), &
       wfc, lda, wfc, lda, (0.d0, 0.d0), overlap, lda)
    
  CALL mp_sum(  overlap, intra_bgrp_comm )

  ! find O^-.5
  !
  CALL cdiagh (lda, overlap, lda, e, work)
  DO i = 1, lda
     IF(ABS(e(i)).lt.1.d-10) THEN
        ierr = 1
        RETURN
     ELSE
        e (i) = 1.d0/dsqrt(e(i))
     END IF
  ENDDO
  
  overlap = (0.d0,0.d0)
  
  DO i = 1, lda
     DO j = 1, lda
        overlap (i, j) = (0.d0,0.d0)
        DO k = 1, lda
           overlap (i, j) = overlap (i, j) + e(k)*work(i, k)*DCONJG(work (j, k) )
        ENDDO
     ENDDO
  ENDDO
  !
  ! trasform wfs O^-.5 psi
  !
  wfc_ortho(:,:) = (0.d0,0.d0)
  call ZGEMM('N', 'N', lda, ldb, lda, (1.d0,0.d0), overlap, lda, &
       wfc, lda, (0.d0,0.d0), wfc_ortho, lda)
 
  wfc(:,:) = wfc_ortho(:,:)
         
  DEALLOCATE (overlap)
  DEALLOCATE (work)
  DEALLOCATE (e)
  DEALLOCATE (wfc_ortho)
  
  RETURN

END SUBROUTINE

SUBROUTINE check_ortho(lda,ldb,wfc)
!This subroutine checks orthogonality of wfs. Created for debug purposes.

  USE kinds,      ONLY : DP
  USE io_global,  ONLY : stdout
  USE noncollin_module, ONLY : noncolin, npol
  
  implicit none
  INTEGER, intent(in) :: lda,ldb
  COMPLEX(DP), intent(in) :: wfc(lda,ldb)
  
  INTEGER :: i,j,k
  COMPLEX(DP), allocatable :: overlap(:,:)
  
  
  ALLOCATE (overlap( lda , lda))    
  overlap = (0.d0,0.d0)
  
  !
  ! calculate overlap matrix
  !
  CALL ZGEMM ('n', 'c', lda, lda, ldb, (1.d0,0.d0), &
       wfc, lda, wfc, lda, (0.d0,0.d0), overlap, lda)

  write(stdout,'(5x,a45,2i5)') 'check_ortho for wavefunction with dimentions ', lda,ldb
  do i=1,lda
     write(stdout,'(5x,8f8.4)') (dreal(overlap(i,j)),j=1,lda)
  end do
  write(stdout,'(5x,a18)') 'end of check_ortho'

  DEALLOCATE (overlap)
  
  RETURN

END SUBROUTINE
