/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2011 Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!--------------------------------------------------------------------------
!
MODULE start_k
  !
  ! ... Basic variables for k-point generations, as read from input
  !
  USE kinds,      ONLY : DP
  USE cell_base,  ONLY : bg
  !
  SAVE
  !
  ! ... uniform k-point grid parameters
  !
  INTEGER :: &
       nk1, nk2, nk3,   &! the special-point grid
       k1, k2, k3        ! the offset from the origin
  !
  !
  ! ... k points and weights, read from input, if any
  !
  INTEGER :: nks_start=0                 ! number of  k points
  REAL(DP), ALLOCATABLE :: wk_start(:)   ! weights of k points
  REAL(DP), ALLOCATABLE :: xk_start(:,:) ! coordinates of k points
  
  CONTAINS
 
    SUBROUTINE init_start_k ( nk1_, nk2_, nk3_, k1_, k2_, k3_, &
                              k_points, nk_, xk_, wk_ ) 
       !
       ! initialize the grid of k points 
       !
       INTEGER, INTENT (IN) :: nk1_, nk2_, nk3_, k1_, k2_, k3_, nk_
       CHARACTER(LEN=*), INTENT (IN) :: k_points
       REAL(dp),INTENT (IN) :: xk_(3,nk_), wk_(nk_)
       !
       LOGICAL :: done
       !
       ! variables for automatic grid
       !
       nk1 = 0; nk2 = 0; nk3 = 0; k1 = 0; k2 = 0; k3 = 0
       done = reset_grid ( nk1_, nk2_, nk3_, k1_, k2_, k3_ )
       IF ( k_points == 'automatic' .AND. .not. done ) &
          CALL errore ('init_start_k','automatic k-points and nk*=0?',1)
       !
       ! variables for manual grid
       !
       IF ( k_points == 'gamma' ) THEN
          nks_start = 1
       ELSE
          nks_start = nk_
       END IF
       !
       IF ( nks_start > 0) THEN
          IF ( .NOT. ALLOCATED (xk_start) ) ALLOCATE ( xk_start(3,nks_start) )
          IF ( .NOT. ALLOCATED (wk_start) ) ALLOCATE ( wk_start(nks_start) )
          !
          ! k-points in crystal axis: transform to cartesian (in units 2pi/a)
          ! BEWARE: reciprocal axis bg NEEDED, must have been initialized
          !
          IF ( k_points == 'crystal' ) CALL cryst_to_cart(nk_, xk_, bg, 1)
          !
          IF ( k_points == 'gamma' ) THEN
            xk_start(:,1) = 0.0_dp
            wk_start(1)   = 1.0_dp
          ELSE
            xk_start(:,:) = xk_(:,1:nk_)
            wk_start(:)   = wk_(1:nk_)
          ENDIF
       END IF
       !
    END SUBROUTINE init_start_k 
    !
    LOGICAL FUNCTION reset_grid ( nk1_, nk2_, nk3_, k1_, k2_, k3_ ) 
       !
       ! reset the automatic grid to new values if these are > 0
       !
       INTEGER, INTENT (IN) :: nk1_, nk2_, nk3_, k1_, k2_, k3_
       !
       IF ( nk1_ > 0 ) nk1 = nk1_
       IF ( nk2_ > 0 ) nk2 = nk2_
       IF ( nk3_ > 0 ) nk3 = nk3_
       IF (  k1_ > 0 )  k1 = k1_
       IF (  k2_ > 0 )  k2 = k2_
       IF (  k3_ > 0 )  k3 = k3_
       !
       reset_grid = (nk1_*nk2_*nk3_ > 0)
       !
    END FUNCTION reset_grid

END MODULE start_k
