/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

! Copyright (C) 2008 Dmitry Korotin dmitry@korotin.name
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!


subroutine wannier_check()
  use io_global,  only : stdout
  use kinds, only : DP 
  use klist, only : nks, nkstot
  use ions_base, only : nat, ityp, atm,tau
  use wvfct, only: nbnd
  USE basis, only: natomwfc
  use wannier_new, only: nwan, wan_in, use_energy_int
  use lsda_mod, only: nspin
  USE control_flags,    ONLY : gamma_only
  USE uspp_param,       ONLY : upf 

  implicit none
  integer :: nwfc, lmax_wfc, na, nt, n, l, m, i, iwan, ispin

  ! number of k points in this pool .ne. total number of k points
  if (nks.ne.nkstot) call errore ('wannier_check', 'not implemented 1', 1)
  
  if ( gamma_only ) call errore ('wannier_check', 'gamma_only calculation not implemented', 1) 

  ! here we will write to stdout source of wannier functions (atomic functions from which wannier are generated) 
  do ispin=1, nspin
     !
     write(stdout,'(5x,a4,i2)') 'Spin',ispin
     do iwan=1,nwan
        write(stdout,'(7x,"Wannier #",i3," centered on atom ",a3," (position ",3f8.5," )")') &
             iwan, atm(ityp(wan_in(iwan,ispin)%iatom)), (tau(i,wan_in(iwan,ispin)%iatom),i=1,3)
        
        if( use_energy_int) then
           write(stdout,'(9x,"Bands for generation: from",f6.3," to",f6.3)') &
                wan_in(iwan,ispin)%bands_from,wan_in(iwan,ispin)%bands_to
        else 
           write(stdout,'(9x,"Bands for generation: from",i4," to",i4)') &
                INT(wan_in(iwan,ispin)%bands_from),INT(wan_in(iwan,ispin)%bands_to)
        end if
        
        write(stdout,'(9x,a31)') 'Trial wavefunction ingredients:'
        do i=1,wan_in(iwan,ispin)%ning
           nwfc=0 
           lmax_wfc = 0 
           write(stdout,'(10x,f12.10," of l=",i1,", m=",i1)') &
                wan_in(iwan,ispin)%ing(i)%c, wan_in(iwan,ispin)%ing(i)%l, wan_in(iwan,ispin)%ing(i)%m
           
           ! now we shoud associate every ingridient of trial wavefunction with atomic orbital
           ! it will be done only once - for future using in wannier_proj
           DO na = 1, nat
              nt = ityp (na)
              DO n = 1, upf(nt)%nwfc
                 IF (upf(nt)%oc (n) >= 0.d0) THEN
                    l = upf(nt)%lchi (n)
                    lmax_wfc = max (lmax_wfc, l ) 
                    DO m=1, 2*l+1          
                       nwfc=nwfc+1 
                       ! the most important part
                       if ( &
                            (na == wan_in(iwan,ispin)%iatom) .AND. &
                            (l == wan_in(iwan,ispin)%ing(i)%l) .AND. &
                            (m == wan_in(iwan,ispin)%ing(i)%m) ) &
                            wan_in(iwan,ispin)%ing(i)%iatomwfc = nwfc
                    enddo
                 endif
              enddo
           enddo
           
        end do ! ingredients
     end do ! iwannier
  end do !ispin
  
  ! do iwan=1,nwan
  !   write(stdout,'(7x,"Wannier #",i3," atomic wavefunction", i3)') iwan, wan_in(iwan,1)%ing(1)%iatomwfc
 ! end do ! iwannier
     
  if (lmax_wfc > 3) call errore ('wannier_check', 'l > 3 not yet implemented', 1) 
  if (nwfc /= natomwfc) call errore ('wannier_check', 'wrong # of atomic wfcs?', 1)
  if (nwan > nbnd ) call errore( 'wannier_check','too few bands', nwan-nbnd)

  return

end subroutine wannier_check
