/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2003 A. Smogunov
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
subroutine gramsh (n, nvec, nstart, nfinish,         &
                    psibase, psiprob, ndim, epsproj)
!
! This routine orthogonalizes a set of vectors with respect
! to the basis set and supplies the latter. It uses the
! Gram-Schmidt method.
!
  USE kinds, only : DP
  implicit none
  integer ::         &
     n,              &  ! input: physical dimension
     nvec,           &  ! input: number of vectors
     nstart,         &  ! input: first vector to orthogonalize
     nfinish,        &  ! input: last vector to orthogonalize
     ndim,           &  ! inp/out: dimension of psibase old/new
     ivec,           &  ! counter on vectors
     ivecp              ! counter on vectors
  real(DP) ::   &
     epsproj,        &  ! accuracy
     norm,           &  ! the norm of a vector
     ddot               ! to compute the dot product of two vectors
  real(DP), parameter :: eps=1.d-8
  complex(DP) :: &
     psibase(n,n),    & ! i/o:basis vector set
     psiprob(n,nvec), & ! i/o:vectors to be orthog. and added to psibas
     zdotc              ! to compute scalar products
  complex(DP), allocatable ::  &
     ps(:)                        ! the scalar products

  allocate( ps( n ) )

  if (ndim.eq.n) return

  do ivec = nstart, nfinish
   if(ndim.lt.n) then
!
! To find orthogonal to psibase projection of psiprob
!
     do ivecp=1, ndim
       ps(ivecp)=zdotc(n,psibase(1,ivecp),1,psiprob(1,ivec),1)
     enddo
     do ivecp=1,ndim
       call zaxpy (n,-ps(ivecp),psibase(1,ivecp),1,psiprob(1,ivec),1)
     enddo
! and its norm
     norm=ddot(2*n,psiprob(1,ivec),1,psiprob(1,ivec),1)
!
! adding (or not) psiprob to psibase
!
     if (norm.le.-eps) then
       print*,'norma = ',norm,ivec
       call errore ('gramsh',' negative norm in S ',1)
     endif
     if (abs(norm).gt.epsproj) then
       ndim=ndim+1
       if (ndim.eq.n) then
          psibase=(0.d0,0.d0)
          do ivecp=1, n
             psibase(ivecp,ivecp)=(1.d0,0.d0)
          enddo
       else
         norm = 1.d0/sqrt(norm)
         call dscal( 2*n,norm,psiprob(1,ivec),1 )
         call dcopy(2*n,psiprob(1,ivec),1,psibase(1,ndim),1)
       endif
     endif
   endif
  enddo

  deallocate(ps)

  return
end subroutine gramsh
