/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!-----------------------------------------------------------------------
!
FUNCTION lr_dot(x,y)
  !---------------------------------------------------------------------
  ! Brent Walker, ICTP, 2004
  !---------------------------------------------------------------------
  ! ... wrapper for PWSCF linear response inner product routines
  ! ... sums over the bands
  ! ... call for each k-point with arguments:
  ! ... call lr_dot(npw_k(ik),evc1(1,1,ik,1),1,evc1(1,1,ik,2),1)
  !---------------------------------------------------------------------
  ! Modified by Osman Baris Malcioglu (2009)
  !
  USE io_global,            ONLY : stdout
  USE kinds,                ONLY : dp
  USE klist,                ONLY : nks
  !use lr_variables,         only : npw_k
  USE realus,               ONLY : npw_k
  USE lsda_mod,             ONLY : nspin
  USE wvfct,                ONLY : npwx,nbnd,wg
  USE control_flags,        ONLY : gamma_only
  USE gvect,                ONLY : gstart
  USE mp,                   ONLY : mp_sum
  USE mp_global,            ONLY : inter_pool_comm, intra_bgrp_comm
  USE lr_variables,   ONLY : lr_verbosity
   USE io_global,      ONLY : stdout
 !
  IMPLICIT NONE
  !
  COMPLEX(kind=dp) :: x(npwx,nbnd,nks),y(npwx,nbnd,nks)
  COMPLEX(kind=dp) :: lr_dot
  COMPLEX(kind=dp) :: temp_k
  real(kind=dp) :: temp_gamma
  real(kind=dp) :: degspin
  INTEGER :: ibnd
  INTEGER :: ik
  real(kind=dp), EXTERNAL    :: DDOT
  COMPLEX(kind=dp), EXTERNAL :: zdotc
  !
  IF (lr_verbosity > 5) THEN
     WRITE(stdout,'("<lr_dot>")')
  ENDIF
  CALL start_clock ('lr_dot')
  !
  lr_dot=(0.0d0,0.0d0)
  !
  temp_gamma=0.0d0
  temp_k=(0.0d0,0.0d0)
  !
  degspin=2.0d0
  IF(nspin==2) degspin=1.0d0
  !
  IF(gamma_only) THEN
     CALL lr_dot_gamma()
     lr_dot=cmplx(temp_gamma,0.0d0,dp)
  ELSE
     CALL lr_dot_k()
     lr_dot=temp_k
  ENDIF
  !
  lr_dot=lr_dot/degspin
  !
  IF (lr_verbosity > 5) WRITE(stdout,'("<end of lr_dot>")')
  CALL stop_clock ('lr_dot')
  !
  RETURN
  !
CONTAINS
  !
  SUBROUTINE lr_dot_gamma
    !
    DO ibnd=1,nbnd
       !
       temp_gamma = temp_gamma + 2.D0*wg(ibnd,1)*DDOT(2*npw_k(1),x(:,ibnd,1),1,y(:,ibnd,1),1)
       IF (gstart==2) temp_gamma = temp_gamma - wg(ibnd,1)*dble(x(1,ibnd,1))*dble(y(1,ibnd,1))
       !
    ENDDO
    !
    !call reduce(1,temp_gamma)
    CALL mp_sum(temp_gamma, intra_bgrp_comm)
    !
    RETURN
  END SUBROUTINE lr_dot_gamma
  !
  SUBROUTINE lr_dot_k
    !
    DO ik=1,nks
       DO ibnd=1,nbnd
          !
          temp_k=temp_k+wg(ibnd,ik)*zdotc(npw_k(ik),x(1,ibnd,ik),1,y(1,ibnd,ik),1)
          !
       ENDDO
    ENDDO
    !call poolreduce(2,temp_k)
    CALL mp_sum(temp_k,inter_pool_comm)
    !call reduce(2,temp_k)
    CALL mp_sum(temp_k, intra_bgrp_comm)
    !
    RETURN
  END SUBROUTINE lr_dot_k
  !
END FUNCTION lr_dot
!-----------------------------------------------------------------------
