/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2004 PWSCF group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!
!---------------------------------------------------------------
SUBROUTINE all_electron(ild,ic)
  !---------------------------------------------------------------
  !
  !  this routine is a driver to an all-electron calculation
  !  with the parameters given in input
  !
  !
  USE kinds, ONLY : DP
  USE radial_grids, ONLY: ndmx
  USE ld1inc, ONLY: isic, grid, zeta, rho, enne, vpot, vxt, enl, &
                     deld, encl, etot, ecxc, evxt, ehrt, epseu, ekin, &
                     vnl, vh, lsd, nspin, nlcc, vdw, nn, ll, oc, nwf, &
                     zed, zval, vxc, exc, excgga, v0, verbosity, &
                     relpert, evel, edar, eso, vsic, vsicnew, vhn1, egc
  IMPLICIT NONE

  INTEGER, INTENT(in) :: ic   ! counter on configurations
  LOGICAL :: ild    ! if true compute log der
  !
  !    compute an initial estimate of the potential
  !
  CALL starting_potential (ndmx, grid%mesh, zval, zed, nwf, oc, nn, ll,&
                           grid%r,enl, v0, vxt, vpot, enne, nspin )
  !
  ! allocate variables for SIC, if needed
  !
  IF (isic /= 0) THEN
     ALLOCATE(vsic(ndmx,nwf), vsicnew(ndmx), vhn1(ndmx), egc(ndmx))
     vsic=0.0_dp
  ENDIF
  !
  !     solve the eigenvalue self-consistent equation
  !
  CALL scf(ic)
  !
  !   compute relativistic corrections to the eigenvalues
  !
  IF ( relpert ) CALL compute_relpert(evel,edar,eso)
  !
  !  compute total energy
  !
  CALL elsd (zed,grid,rho,vxt,vh,vxc,exc,excgga,nwf,nspin,enl,oc,    &
             etot,ekin,encl,ehrt,ecxc,evxt)
  !
  IF (verbosity=='high') CALL elsd_highv(ic)
  !
  !   add sic correction if needed
  !
  IF(isic /= 0) CALL esic
  !
  !   print results
  !
  CALL write_results
  !
  !  compute logarithmic derivative
  !
  IF (deld > 0.0_DP .and. ild) CALL lderiv
  !
  ! compute C6 coefficient if required
  !
  IF (vdw) THEN
     CALL c6_tfvw ( grid%mesh, zed, grid, rho(1,1) )
     CALL c6_dft  ( grid%mesh, zed, grid )
  ENDIF
  !
  IF (isic /= 0) THEN
     DEALLOCATE(egc, vhn1, vsicnew, vsic)
  ENDIF
  !
  RETURN
  !
END SUBROUTINE all_electron
