/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2004 PWSCF group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!
!----------------------------------------------------------------------
subroutine integrate_inward(e,mesh,ndm,grid,f,y,c,el,ik,nstart)
  !----------------------------------------------------------------------
  !
  !     this subroutine integrate inward the schroedinger equation
  !     only local potential allowed
  !
  use kinds, only : DP
  use radial_grids, only: radial_grid_type
  implicit none
  type(radial_grid_type), intent(in):: grid
  integer :: &
       mesh,  &    ! size of radial mesh
       ndm,   &    ! maximum radial mesh
       ik          ! the matching point

  real(DP) :: &
       e,       &  ! output eigenvalue
       f(mesh), &  ! the function defining the equation
       y(mesh), &  ! the output solution
       c(mesh),el(mesh) ! auxiliary space

  real(DP) :: &
       rstart,  &  ! the starting r of the inward integration
       di,      &  ! auxiliary for integration  
       expn        ! exponential for tail of wavefunction

  integer :: &
       nstart, &   ! the starting point of inward integration
       n           ! counter on mesh points

  !
  !     prepare inward integration
  !     charlotte froese can j phys 41,1895(1963)
  !
  !     start at  min( rmax, 10*rmatch )
  !
  nstart=mesh
  if (mesh .ne. grid%mesh) call errore('integrate_inward','mesh dimention not as expected',1)
  rstart=10.0_dp*grid%r(ik)
  if (rstart.lt.grid%r(mesh)) then
     do n=ik,mesh
        nstart=n
        if(grid%r(n).ge.rstart) go to 100
     enddo
100  if (mod(nstart,2) == 0) nstart=nstart+1
  endif
  !
  !  set up a, l, and c vectors
  !
  n=ik+1
  el(n)=10.0_dp*f(n)-12.0_dp
  c(n)=-f(ik)*y(ik)
  do n=ik+2,nstart
     di=10.0_dp*f(n)-12.0_dp
     el(n)=di-f(n)*f(n-1)/el(n-1)
     c(n)=-c(n-1)*f(n-1)/el(n-1)
  enddo
  !
  !  start inward integration by the froese's tail procedure
  !
  n=nstart-1
  expn=exp(-sqrt(12.0_dp*abs(1.0_dp-f(n))))
  y(n)=c(n)/(el(n)+f(nstart)*expn)
  y(nstart)=expn*y(n)
  !
  !    and integrate inward
  !
  do n=nstart-2,ik+1,-1
     y(n)=(c(n)-f(n+1)*y(n+1))/el(n)
  enddo

  return
end subroutine integrate_inward
