/* Copyright (C) 1991-2023 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <https://www.gnu.org/licenses/>.  */


/* This header is separate from features.h so that the compiler can
   include it implicitly at the start of every compilation.  It must
   not itself include <features.h> or any other header that includes
   <features.h> because the implicit include comes before any feature
   test macros that may be defined in a source file before it first
   explicitly includes a system header.  GCC knows the name of this
   header in order to preinclude it.  */

/* glibc's intent is to support the IEC 559 math functionality, real
   and complex.  If the GCC (4.9 and later) predefined macros
   specifying compiler intent are available, use them to determine
   whether the overall intent is to support these features; otherwise,
   presume an older compiler has intent to support these features and
   define these macros by default.  */



/* wchar_t uses Unicode 10.0.0.  Version 10.0 of the Unicode Standard is
   synchronized with ISO/IEC 10646:2017, fifth edition, plus
   the following additions from Amendment 1 to the fifth edition:
   - 56 emoji characters
   - 285 hentaigana
   - 3 additional Zanabazar Square characters */

!
! Copyright (C) 2010 Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!

SUBROUTINE set_psi_in(ik,l,j,e,psi_out,psi_out_rel)
  !
  !  This subroutine calculates the all electron wavefunction psi at the
  !  input energy e
  !
  USE kinds, ONLY : dp
  USE radial_grids, ONLY : ndmx
  USE ld1inc, ONLY : grid, rel, zed, vpot
  IMPLICIT NONE
  INTEGER :: l, ik    ! input: angular momentum and index of the cut-off radius
  REAL(DP) :: e, j    ! input: energy and total angular momentum
  REAL(DP) :: psi_out(ndmx) ! output: the function psi.
  REAL(DP) :: psi_out_rel(ndmx) ! output: the function psi (small component).
  REAL(DP) :: psi_dir(ndmx,2) ! auxiliary function.
  REAL(DP) :: ze2, jnor, thrdum=0.0_dp
  INTEGER  :: n, i, nstop

  psi_out_rel=0.0_DP
  IF (rel == 1) THEN
     n =  grid%mesh
     CALL lschps (3, zed, thrdum, grid, n, 1, l, e, vpot, psi_out, nstop )
  ELSEIF (rel == 2) THEN
     CALL dir_outward (ndmx, grid%mesh, l, j, e, grid%dx, psi_dir, &
          grid%r, grid%rab, vpot )
     psi_out(:)=psi_dir(:,1)
     psi_out_rel(:)=psi_dir(:,2)
  ELSE
     ze2=-zed*2.0_dp
     CALL intref(l,e,grid%mesh,grid,vpot,ze2,psi_out)
  ENDIF
  !
  !    fix arbitrarily the norm at the cut-off radius equal to (about) 0.5**2
  !
  jnor=0.0_dp
  DO n=1,ik
     jnor = jnor + grid%dx*grid%r(n)*psi_out(n)**2
  ENDDO
  jnor = sqrt(jnor)
  DO n=1,grid%mesh
     psi_out(n)=psi_out(n)*0.5_dp/jnor
  ENDDO
  IF (rel==2) THEN
     DO n=1,grid%mesh
        psi_out_rel(n)=psi_out_rel(n)*0.5_dp/jnor
     ENDDO
  ENDIF
  !
  !  Set the wavefunction to zero if it diverges too much. In any case
  !  this part of the scattering wavefunction is not used and generates 
  !  noise in the code.
  !
  Do n=ik+1,grid%mesh
     IF (ABS(psi_out(n))>1.d9) THEN
        DO i=n,grid%mesh
           psi_out(i)=0.0_DP
           IF (rel==2) psi_out_rel(i)=0.0_DP
        ENDDO
     ENDIF
  ENDDO 

  RETURN
END SUBROUTINE set_psi_in

